#include <iostream>
#include <fstream>
#include <sstream>
#include "kernel.h"

void DataGrid::read_from_file(const std::string& input_file) {
    std::ifstream file(input_file);
    std::string str;

    // header
    const std::string delimiter = ":";
    while (std::getline(file, str)) {
        if (!isdigit(str[0])) {
            std::string k = str.substr(0, str.find(delimiter));
            int v = stoi(str.substr(str.find(delimiter)+2, str.size()));
            metadata.insert({k, v});
        }
        else {
            break;
        }
    }
    if (metadata["rows"] <= 0 || metadata["cols"] <= 0) {
         throw std::invalid_argument("Invalid grid size");
    }

    // body
    do {
        std::stringstream ss(str);
        std::istringstream split(str);
        std::vector<int> line;
        std::string svalue;
        while (std::getline (split, svalue, ' ')) {
           line.push_back(stoi(svalue));
        }
        data.push_back(line); // TODO: use reference
        if ((int) line.size() != metadata["cols"]) {
            throw std::invalid_argument("Inconsistency in cols");
        }
    }  while (std::getline(file, str));

    if ((int) data.size() != metadata["rows"]) {
        throw std::invalid_argument("Inconsistency in rows");
    }
}

std::ostream &operator<<(std::ostream &out, const DataGrid &grid)
{
    if (grid.metadata.size() < 1) {
        out << "Grid not initialized" << std::endl;
        return out;
    }
   out << "Grid initialized by:" << std::endl;

   //std::map<std::string, int>::iterator it;
   //for (it = grid.metadata.begin(); it != grid.metadata.end(); it++) {
   for (auto const& i : grid.metadata) {
        out << " " << i.first << ": " << i.second << std::endl;
   }
   out << "Data ";
   if (grid.data.size() < 1)
       out << "not ";
   out << "filled" << std::endl;

   return out;
}

DataGrid &DataGrid::apply_kernel(KernelCount &kernel) {
    DataGrid *output = new DataGrid(metadata["rows"], metadata["cols"]);

    for (int i = 0; i < metadata["rows"]; i++) {
        for (int j = 0; j < metadata["cols"]; j++) {
            kernel.reset();
            for (int l = 0; l < kernel.metadata["rows"]; l++) {
                for(int k = 0; k < kernel.metadata["cols"]; k++) {
                    int i_l = i + (l - kernel.size());
                    int j_k = j + (k - kernel.size());
                    if (i_l > -1 && j_k > -1 &&
                        i_l < metadata["rows"] &&
                        j_k < metadata["cols"]) {
                        kernel.data[l][k] = data[i_l][j_k];
                    }
                }
            }
            output->data[i][j] = kernel.calculate();
        }
        if (i > 0 && i % 10 == 0) {
            int perc = (int) ((i / (float) metadata["rows"]) * 100.);
            std::cout << perc << "%\r" << std::flush;
        }
    }
    std::cout << "100%" << std::endl;
    return *output;
}

void KernelGrid::reset() {
    for (int i = 0; i < metadata["rows"]; i++) {
        for (int j = 0; j < metadata["cols"]; j++) {
            data[i][j] = nullval;
        }
    }
}

int KernelGrid::size() {
    return (int) metadata["rows"] / 2;
}

int KernelCount::calculate() {
    int count = 0;
    for (std::size_t i = 0; i < data.size(); i++) {
        for (std::size_t j = 0; j < data.size(); j++) {
            if (data[i][j] != -1) {
                count += 1;
            }
        }
    }

    return count;

    return 1;
}

std::string Grid::stats() {
    int minv = nullval;
    int maxv = nullval;
    for (std::size_t i = 0; i < data.size(); i++) {
        for (std::size_t j = 0; j < data[i].size(); j++) {
             int val = data[i][j];
             if (minv == nullval || minv > val)
                 minv = val;
             if (maxv == nullval || maxv < val)
                 maxv = val;
        }
    }
    return "min: " + std::to_string(minv) + "; max: " + std::to_string(maxv);
}
